//! `Hoverable` is a property to hover over an item per mouse

use super::framable::Framable;
use crate::prelude::*;
use anyhow::Result;

use std::sync::{
    atomic::{AtomicBool, AtomicI32, Ordering::SeqCst},
    Arc,
};

#[cfg(feature = "audio")]
use std::sync::RwLock;

/// `Hoverable` gives the ability to execute a closure when the mouse hovers over the item
pub struct Hoverable {
    framable: Arc<Framable>,

    #[cfg(feature = "audio")]
    audible: RwLock<Option<Arc<Audible>>>,
    hovered: AtomicBool,

    ui_layer: AtomicI32,

    hover_changed_executable: Arc<Executable<bool>>,
    on_hover_executable: Arc<Executable<bool>>,
}

impl Hoverable {
    /// Factory method for `Hoverable`, returns `Arc<Hoverable>`.
    ///
    /// # Arguments
    ///
    /// * `framable` is a `Arc<Framable>` instance
    pub fn new(
        framable: Arc<Framable>,
        hover_changed_executable: Arc<Executable<bool>>,
        on_hover_executable: Arc<Executable<bool>>,
    ) -> Arc<Self> {
        Arc::new(Hoverable {
            framable,

            hovered: AtomicBool::new(false),
            ui_layer: AtomicI32::new(0),

            hover_changed_executable,
            on_hover_executable,

            #[cfg(feature = "audio")]
            audible: RwLock::new(None),
        })
    }

    /// Add method
    ///
    /// # Arguments
    ///
    /// * `hoverable` is a `&Arc<Hoverable>` instance that is going to be added
    pub fn add(self: &Arc<Self>) -> Result<()> {
        self.framable
            .gui_handler()
            .add_hoverable(self.ui_layer.load(SeqCst), self.clone())?;

        Ok(())
    }

    /// Delete method, has to be explicitly called, otherwise it will remain in memory.
    ///
    /// # Arguments
    ///
    /// * `hoverable` is a `&Arc<Hoverable>` instance that is going to be deleted
    pub fn delete(self: &Arc<Self>) -> Result<()> {
        self.framable
            .gui_handler()
            .delete_hoverable(self.ui_layer.load(SeqCst), self)?;
        self.set_hovered(false)?;

        Ok(())
    }

    pub fn set_ui_layer(&self, ui_layer: i32) {
        self.ui_layer.store(ui_layer, SeqCst);
    }

    #[cfg(feature = "audio")]
    pub fn set_audible(&self, audible: Option<Arc<Audible>>) {
        *self.audible.write().unwrap() = audible;
    }

    /// Returns true if x and y are inside the bounds of the given framable,
    /// otherwise false. Generally used by the `GuiHandler`.
    ///
    /// # Arguments
    ///
    /// * `x` is the x value to check
    /// * `y` is the y value to check
    pub fn is_hovered(&self, x: i32, y: i32) -> bool {
        (x > self.framable.left())
            && (x < self.framable.right())
            && (y > self.framable.top())
            && (y < self.framable.bottom())
    }

    /// Sets the value of hovered and calls the callback if the value changed.
    /// Generally used by the `GuiHandler`.
    ///
    /// # Arguments
    ///
    /// * `hovered` is the new hovered state
    pub fn set_hovered(&self, hovered: bool) -> Result<()> {
        if self.hovered() != hovered {
            self.hovered.store(hovered, SeqCst);

            self.hover_changed_executable.execute(hovered)?;
            self.on_hover_executable.execute(hovered)?;

            #[cfg(feature = "audio")]
            {
                if self.hovered() {
                    if let Some(audible) = self.audible.read().unwrap().as_ref() {
                        audible.play()?;
                    }
                }
            }
        }

        Ok(())
    }

    /// Returns the hovered state. Generally used by the `GuiHandler`.
    pub fn hovered(&self) -> bool {
        self.hovered.load(SeqCst)
    }
}
